@@ -1,7 +1,7 @@
 /* GNU m4 -- A simple macro processor
 
-   Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 2004, 2005, 2006 Free
-   Software Foundation, Inc.
+   Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 2004, 2005, 2006, 2007
+   Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -25,6 +25,10 @@
 #include <limits.h>
 #include <signal.h>
 
+#include "version-etc.h"
+
+#define AUTHORS "Rene' Seindal"
+
 static void usage (int);
 
 /* Enable sync output for /lib/cpp (-s).  */
@@ -48,6 +52,9 @@ int max_debug_argument_length = 0;
 /* Suppress warnings about missing arguments.  */
 int suppress_warnings = 0;
 
+/* If true, then warnings affect exit status.  */
+static bool fatal_warnings = false;
+
 /* If not zero, then value of exit status for warning diagnostics.  */
 int warning_status = 0;
 
@@ -62,11 +69,15 @@ const char *user_word_regexp = "";
 /* The name this program was run with. */
 const char *program_name;
 
+/* Global catchall for any errors that should affect final error status, but
+   where we try to continue execution in the meantime.  */
+int retcode;
+
 struct macro_definition
 {
   struct macro_definition *next;
-  int code;			/* D, U or t */
-  const char *macro;
+  int code;			/* D, U, s, t, or '\1' */
+  const char *arg;
 };
 typedef struct macro_definition macro_definition;
 
@@ -83,6 +94,8 @@ m4_error (int status, int errnum, const char *format, ...)
   va_start (args, format);
   verror_at_line (status, errnum, current_line ? current_file : NULL,
 		  current_line, format, args);
+  if (fatal_warnings && ! retcode)
+    retcode = EXIT_FAILURE;
 }
 
 /*-------------------------------.
@@ -96,6 +109,8 @@ m4_error_at_line (int status, int errnum, const char *file, int line,
   va_list args;
   va_start (args, format);
   verror_at_line (status, errnum, line ? file : NULL, line, format, args);
+  if (fatal_warnings && ! retcode)
+    retcode = EXIT_FAILURE;
 }
 
 #ifdef USE_STACKOVF
@@ -138,10 +153,16 @@ for short options too.\n\
 Operation modes:\n\
       --help                   display this help and exit\n\
       --version                output version information and exit\n\
-  -E, --fatal-warnings         stop execution after first warning\n\
+", stdout);
+      fputs ("\
+  -E, --fatal-warnings         once: warnings become errors, twice: stop\n\
+                               execution at first error\n\
   -i, --interactive            unbuffer output, ignore interrupts\n\
   -P, --prefix-builtins        force a `m4_' prefix to all builtins\n\
   -Q, --quiet, --silent        suppress some warnings for builtins\n\
+      --warn-macro-sequence[=REGEXP]\n\
+                               warn if macro definition matches REGEXP,\n\
+                               default \\$\\({[^}]*}\\|[0-9][0-9]+\\)\n\
 ", stdout);
 #ifdef ENABLE_CHANGEWORD
       fputs ("\
@@ -204,9 +225,6 @@ mismatch, or whatever value was passed to the m4exit macro.\n\
 ", stdout);
       printf ("\nReport bugs to <%s>.\n", PACKAGE_BUGREPORT);
     }
-
-  if (close_stream (stdout) != 0)
-    M4ERROR ((EXIT_FAILURE, errno, "write error"));
   exit (status);
 }
 
@@ -220,6 +238,7 @@ enum
 {
   DEBUGFILE_OPTION = CHAR_MAX + 1,	/* no short opt */
   DIVERSIONS_OPTION,			/* not quite -N, because of message */
+  WARN_MACRO_SEQUENCE_OPTION,		/* no short opt */
 
   HELP_OPTION,				/* no short opt */
   VERSION_OPTION			/* no short opt */
@@ -249,6 +268,7 @@ static const struct option long_options[] =
 
   {"debugfile", required_argument, NULL, DEBUGFILE_OPTION},
   {"diversions", required_argument, NULL, DIVERSIONS_OPTION},
+  {"warn-macro-sequence", optional_argument, NULL, WARN_MACRO_SEQUENCE_OPTION},
 
   {"help", no_argument, NULL, HELP_OPTION},
   {"version", no_argument, NULL, VERSION_OPTION},
@@ -256,14 +276,49 @@ static const struct option long_options[] =
   { NULL, 0, NULL, 0 },
 };
 
-/* Global catchall for any errors that should affect final error status, but
-   where we try to continue execution in the meantime.  */
-int retcode;
+/* Process a command line file NAME, and return true only if it was
+   stdin.  */
+static bool
+process_file (const char *name)
+{
+  bool result = false;
+  if (strcmp (name, "-") == 0)
+    {
+      /* If stdin is a terminal, we want to allow 'm4 - file -'
+	 to read input from stdin twice, like GNU cat.  Besides,
+	 there is no point closing stdin before wrapped text, to
+	 minimize bugs in syscmd called from wrapped text.  */
+      push_file (stdin, "stdin", false);
+      result = true;
+    }
+  else
+    {
+      char *full_name;
+      FILE *fp = m4_path_search (name, &full_name);
+      if (fp == NULL)
+	{
+	  error (0, errno, "%s", name);
+	  /* Set the status to EXIT_FAILURE, even though we
+	     continue to process files after a missing file.  */
+	  retcode = EXIT_FAILURE;
+	  return false;
+	}
+      push_file (fp, full_name, true);
+      free (full_name);
+    }
+  expand_input ();
+  return result;
+}
 
+/* POSIX requires only -D, -U, and -s; and says that the first two
+   must be recognized when interspersed with file names.  Traditional
+   behavior also handles -s between files.  Starting OPTSTRING with
+   '-' forces getopt_long to hand back file names as arguments to opt
+   '\1', rather than reordering the command line.  */
 #ifdef ENABLE_CHANGEWORD
-#define OPTSTRING "B:D:EF:GH:I:L:N:PQR:S:T:U:W:d::eil:o:st:"
+#define OPTSTRING "-B:D:EF:GH:I:L:N:PQR:S:T:U:W:d::eil:o:st:"
 #else
-#define OPTSTRING "B:D:EF:GH:I:L:N:PQR:S:T:U:d::eil:o:st:"
+#define OPTSTRING "-B:D:EF:GH:I:L:N:PQR:S:T:U:d::eil:o:st:"
 #endif
 
 int
@@ -271,18 +326,21 @@ main (int argc, char *const *argv, char *const *envp)
 {
   macro_definition *head;	/* head of deferred argument list */
   macro_definition *tail;
-  macro_definition *new;
+  macro_definition *defn;
   int optchar;			/* option character */
 
   macro_definition *defines;
-  FILE *fp;
-  boolean read_stdin = FALSE;
-  boolean interactive = FALSE;
+  bool read_stdin = false;
+  bool interactive = false;
+  bool seen_file = false;
+  const char *debugfile = NULL;
   const char *frozen_file_to_read = NULL;
   const char *frozen_file_to_write = NULL;
+  const char *macro_sequence = "";
 
   program_name = argv[0];
   retcode = EXIT_SUCCESS;
+  atexit (close_stdout);
 
   include_init ();
   debug_init ();
@@ -294,9 +352,8 @@ main (int argc, char *const *argv, char *const *envp)
 
   head = tail = NULL;
 
-  while (optchar = getopt_long (argc, (char **) argv, OPTSTRING,
-				long_options, NULL),
-	 optchar != EOF)
+  while ((optchar = getopt_long (argc, (char **) argv, OPTSTRING,
+				 long_options, NULL)) != -1)
     switch (optchar)
       {
       default:
@@ -320,25 +377,29 @@ main (int argc, char *const *argv, char *const *envp)
 
       case 'D':
       case 'U':
+      case 's':
       case 't':
-
+      case '\1':
 	/* Arguments that cannot be handled until later are accumulated.  */
 
-	new = (macro_definition *) xmalloc (sizeof (macro_definition));
-	new->code = optchar;
-	new->macro = optarg;
-	new->next = NULL;
+	defn = (macro_definition *) xmalloc (sizeof (macro_definition));
+	defn->code = optchar;
+	defn->arg = optarg;
+	defn->next = NULL;
 
 	if (head == NULL)
-	  head = new;
+	  head = defn;
 	else
-	  tail->next = new;
-	tail = new;
+	  tail->next = defn;
+	tail = defn;
 
 	break;
 
       case 'E':
-	warning_status = EXIT_FAILURE;
+	if (! fatal_warnings)
+	  fatal_warnings = true;
+	else
+	  warning_status = EXIT_FAILURE;
 	break;
 
       case 'F':
@@ -394,7 +455,7 @@ main (int argc, char *const *argv, char *const *envp)
 	error (0, 0, "Warning: `m4 -e' is deprecated, use `-i' instead");
 	/* fall through */
       case 'i':
-	interactive = TRUE;
+	interactive = true;
 	break;
 
       case 'l':
@@ -409,27 +470,21 @@ main (int argc, char *const *argv, char *const *envp)
 	   or later is more widely established, as such a warning
 	   would interfere with all earlier versions of autoconf.  */
       case DEBUGFILE_OPTION:
-	if (!debug_set_output (optarg))
-	  error (0, errno, "%s", optarg);
+	/* Don't call debug_set_output here, as it has side effects.  */
+	debugfile = optarg;
 	break;
 
-      case 's':
-	sync_output = 1;
+      case WARN_MACRO_SEQUENCE_OPTION:
+         /* Don't call set_macro_sequence here, as it can exit.
+            --warn-macro-sequence sets optarg to NULL (which uses the
+            default regexp); --warn-macro-sequence= sets optarg to ""
+            (which disables these warnings).  */
+        macro_sequence = optarg;
 	break;
 
       case VERSION_OPTION:
-	 printf ("%s\n", PACKAGE_STRING);
-	 fputs ("\
-Copyright (C) 2006 Free Software Foundation, Inc.\n\
-This is free software; see the source for copying conditions.  There is NO\n\
-warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
-\n\
-Written by Rene' Seindal.\n\
-", stdout);
-
-	 if (close_stream (stdout) != 0)
-	    M4ERROR ((EXIT_FAILURE, errno, "write error"));
-	 exit (EXIT_SUCCESS);
+	version_etc (stdout, PACKAGE, PACKAGE_NAME, VERSION, AUTHORS, NULL);
+	exit (EXIT_SUCCESS);
 	break;
 
       case HELP_OPTION:
@@ -439,11 +494,14 @@ Written by Rene' Seindal.\n\
 
   defines = head;
 
-  /* Do the basic initialisations.  */
+  /* Do the basic initializations.  */
+  if (debugfile && !debug_set_output (debugfile))
+    M4ERROR ((0, errno, "cannot set debug file `%s'", debugfile));
 
   input_init ();
   output_init ();
   symtab_init ();
+  set_macro_sequence (macro_sequence);
   include_env_init ();
 
   if (frozen_file_to_read)
@@ -451,46 +509,6 @@ Written by Rene' Seindal.\n\
   else
     builtin_init ();
 
-  /* Handle deferred command line macro definitions.  Must come after
-     initialisation of the symbol table.  */
-
-  while (defines != NULL)
-    {
-      macro_definition *next;
-      char *macro_value;
-      symbol *sym;
-
-      switch (defines->code)
-	{
-	case 'D':
-	  macro_value = strchr (defines->macro, '=');
-	  if (macro_value == NULL)
-	    macro_value = "";
-	  else
-	    *macro_value++ = '\0';
-	  define_user_macro (defines->macro, macro_value, SYMBOL_INSERT);
-	  break;
-
-	case 'U':
-	  lookup_symbol (defines->macro, SYMBOL_DELETE);
-	  break;
-
-	case 't':
-	  sym = lookup_symbol (defines->macro, SYMBOL_INSERT);
-	  SYMBOL_TRACED (sym) = TRUE;
-	  break;
-
-	default:
-	  M4ERROR ((warning_status, 0,
-		    "INTERNAL ERROR: bad code in deferred arguments"));
-	  abort ();
-	}
-
-      next = defines->next;
-      free (defines);
-      defines = next;
-    }
-
   /* Interactive mode means unbuffered output, and interrupts ignored.  */
 
   if (interactive)
@@ -499,47 +517,65 @@ Written by Rene' Seindal.\n\
       setbuf (stdout, (char *) NULL);
     }
 
-  /* Handle the various input files.  Each file is pushed on the input,
+  /* Handle deferred command line macro definitions.  Must come after
+     initialization of the symbol table.  */
+
+  while (defines != NULL)
+    {
+      macro_definition *next;
+      symbol *sym;
+
+      switch (defines->code)
+	{
+	case 'D':
+	  {
+	    /* defines->arg is read-only, so we need a copy.  */
+	    char *macro_name = xstrdup (defines->arg);
+	    char *macro_value = strchr (macro_name, '=');
+	    if (macro_value)
+	      *macro_value++ = '\0';
+	    define_user_macro (macro_name, macro_value, SYMBOL_INSERT);
+	    free (macro_name);
+	  }
+	  break;
+
+	case 'U':
+	  lookup_symbol (defines->arg, SYMBOL_DELETE);
+	  break;
+
+	case 't':
+	  sym = lookup_symbol (defines->arg, SYMBOL_INSERT);
+	  SYMBOL_TRACED (sym) = true;
+	  break;
+
+	case 's':
+	  sync_output = 1;
+	  break;
+
+	case '\1':
+	  seen_file = true;
+	  if (process_file (defines->arg))
+	    read_stdin = true;
+	  break;
+
+	default:
+	  M4ERROR ((0, 0, "INTERNAL ERROR: bad code in deferred arguments"));
+	  abort ();
+	}
+
+      next = defines->next;
+      free (defines);
+      defines = next;
+    }
+
+  /* Handle remaining input files.  Each file is pushed on the input,
      and the input read.  Wrapup text is handled separately later.  */
 
-  if (optind == argc)
-    {
-      /* No point closing stdin until after wrapped text is
-	 processed.  */
-      push_file (stdin, "stdin", FALSE);
-      read_stdin = TRUE;
-      expand_input ();
-    }
+  if (optind == argc && !seen_file)
+    read_stdin = process_file ("-");
   else
     for (; optind < argc; optind++)
-      {
-	if (strcmp (argv[optind], "-") == 0)
-	  {
-	    /* If stdin is a terminal, we want to allow 'm4 - file -'
-	       to read input from stdin twice, like GNU cat.  Besides,
-	       there is no point closing stdin before wrapped text, to
-	       minimize bugs in syscmd called from wrapped text.  */
-	    push_file (stdin, "stdin", FALSE);
-	    read_stdin = TRUE;
-	  }
-	else
-	  {
-	    const char *name;
-	    fp = path_search (argv[optind], &name);
-	    if (fp == NULL)
-	      {
-		error (0, errno, "%s", argv[optind]);
-		/* Set the status to EXIT_FAILURE, even though we
-		   continue to process files after a missing file.  */
-		retcode = EXIT_FAILURE;
-		continue;
-	      }
-	    push_file (fp, name, TRUE);
-	    free ((char *) name);
-	  }
-	expand_input ();
-      }
-#undef NEXTARG
+      read_stdin |= process_file (argv[optind]);
 
   /* Now handle wrapup text.  */
 
@@ -550,9 +586,9 @@ Written by Rene' Seindal.\n\
      stream and detect any errors it might have encountered.  Close
      stdin if we read from it, to detect any errors.  */
   debug_set_output (NULL);
-  if (read_stdin && fclose (stdin) == EOF)
+  if (read_stdin && close_stream (stdin) == EOF)
     {
-      M4ERROR ((warning_status, errno, "error reading file"));
+      M4ERROR ((warning_status, errno, "error reading stdin"));
       retcode = EXIT_FAILURE;
     }
 
@@ -563,8 +599,7 @@ Written by Rene' Seindal.\n\
       make_diversion (0);
       undivert_all ();
     }
-
-  if (close_stream (stdout) != 0)
-    M4ERROR ((EXIT_FAILURE, errno, "write error"));
+  output_exit ();
+  free_macro_sequence ();
   exit (retcode);
 }
